from logging import getLogger
from time import monotonic
from typing import Callable


from homeconnect_watcher import VERSION
from homeconnect_watcher.event import HomeConnectEvent


class Metrics:
    def __init__(self, port: int):
        self.logger = getLogger(self.__class__.__name__)
        try:
            from prometheus_client import Counter, Gauge, Info, start_http_server
        except ImportError:
            self.logger.error("Unable to expose prometheus metrics; prometheus_client is not installed.")
            self.logger.error(
                "Please run `pip install prometheus_client` or `pip install 'homeconnect-watcher[prometheus]'`."
            )
            return
        self._start_time = monotonic()
        self._disconnects = Counter("disconnects", "The number of time the connection failed.", ["reason"])
        self._disconnects.labels(reason="timeout")
        self._events = Counter("events", "Number of events.", ["appliance_id", "event"])
        self._info = Info("version", "Version info.")
        self._info.info({"version": VERSION})
        self._last_event = Gauge("last_event", "Time since last event.")
        self._metric_uptime = Gauge("uptime", "Watcher uptime.")
        self._metric_uptime.set_function(lambda: monotonic() - self._start_time)
        self._n_appliances = Gauge("n_appliances", "The number of known appliances.")
        self._token_refresh = Counter("token_refresh", "The number of times the token was refreshed.")
        self.logger.info(f"Exposing prometheus metrics on port {port}.")
        start_http_server(port)

    def increment_disconnects(self, reason: str) -> None:
        self._disconnects.labels(reason=reason).inc()

    def increment_event_counter(self, event: HomeConnectEvent) -> None:
        self._events.labels(appliance_id=event.appliance_id, event=event.event).inc()

    def increment_token_refresh(self) -> None:
        self._token_refresh.inc()

    def set_last_event(self, function: Callable[[], float]) -> None:
        """
        Set the function for the last event metric.

        This should be a function that returns the time since the last event in seconds, e.g.
            lambda: monotonic() - last_event
        """
        self._last_event.set_function(function)

    def set_n_appliances(self, function: Callable[[], int]) -> None:
        """
        Set the function for the n_appliances metric.

        This should be a function that returns the number of appliances., e.g.
            lambda: len(appliances)
        """
        self._n_appliances.set_function(function)



from dataclasses import dataclass


@dataclass
class Trigger:
    appliance_id: str
    status: bool = False
    settings: bool = False
    active_program: bool = False
    selected_program: bool = False
    interval: bool = False



from asyncio import sleep
from logging import getLogger
from typing import Any, Callable, Type, TypeVar

logger = getLogger(__name__)
F = TypeVar(name="F", bound=Callable[..., Any])


def retry(n_tries: int, delay: float = 1, exceptions: Type[Exception] | tuple[Type[Exception]] = Exception):
    def func_wrapper(f: F) -> F:
        async def wrapper(*args, **kwargs):
            for i_try in range(n_tries):
                try:
                    return await f(*args, **kwargs)
                except exceptions as exc:
                    logger.debug(f"Function {f.__name__} failed with exception {exc} in try {i_try} out of {n_tries}.")
                    if i_try + 1 == n_tries:
                        raise exc
                    else:
                        await sleep(delay)

        return wrapper

    return func_wrapper



from asyncio import TimeoutError, wait_for
from typing import AsyncIterable

from homeconnect_watcher.exceptions import HomeConnectTimeout


async def timeout(iterable: AsyncIterable, duration: int) -> AsyncIterable:
    try:
        iterator = iterable.__aiter__()
        while True:
            yield await wait_for(iterator.__anext__(), timeout=duration)
    except TimeoutError:
        raise HomeConnectTimeout()
    except StopAsyncIteration:
        return



from enum import Enum
from logging import basicConfig, DEBUG, INFO, WARNING


class LogLevel(str, Enum):
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"


def initialize_logging(level: LogLevel) -> None:
    level = WARNING if level == LogLevel.WARNING else INFO if level == LogLevel.INFO else DEBUG
    basicConfig(level=level, format="%(asctime)s %(name)s - %(levelname)s:%(message)s")



from dataclasses import dataclass
from json import dumps, loads
from time import time

from homeconnect_watcher.client.trigger import Trigger


@dataclass
class HomeConnectEvent:
    event: str
    timestamp: int
    appliance_id: str | None = None
    data: dict[str, ...] | None = None
    error: dict[str, ...] | None = None

    @classmethod
    def from_request(cls, request: str, appliance_id: str, response: dict[str, ...]) -> "HomeConnectEvent":
        if "error" in response:
            error = response["error"]
            data = None
        else:
            data = response["data"]
            error = None
        return HomeConnectEvent(
            event=f"{request}-REQUEST",
            timestamp=int(time()),
            appliance_id=appliance_id,
            data=data,
            error=error,
        )

    @classmethod
    def from_stream(cls, stream: bytes) -> "HomeConnectEvent":
        data = {"timestamp": int(time())}
        for line in stream.decode("utf-8").split("\n"):
            if line.startswith("data:") and len(line) > 5:
                data["data"] = loads(line[5:])
            elif line.startswith("event:"):
                data["event"] = line[6:].strip()
            elif line.startswith("id:"):
                data["appliance_id"] = line[3:].strip()
        return cls(**data)

    @classmethod
    def from_string(cls, string: str) -> "HomeConnectEvent":
        return cls(**loads(string))

    @property
    def is_request(self) -> bool:
        return self.event.endswith("-REQUEST")

    @property
    def items(self) -> dict[str, str | None]:
        """Extract the payload into key/value pairs."""
        if self.data is not None:
            if "items" in self.data:  # For STATUS/EVENT/NOTIFY
                return {item["key"]: item["value"] for item in self.data["items"]}
            elif "key" in self.data:  # For CONNECTED/DISCONNECTED
                return {self.data["key"]: self.data["value"]}
            elif "status" in self.data:  # For STATUS-REQUEST
                return self.data["status"]
            elif "settings" in self.data:  # For SETTINGS-REQUEST
                return self.data["settings"]
        elif self.event == "ACTIVE-PROGRAM-REQUEST":
            if self.error_key == "SDK.Error.NoProgramActive":
                return {"BSH.Common.Root.ActiveProgram": None}
            else:
                result = {item["key"]: item["value"] for item in self.data["options"]}
                result["BSH.Common.Root.ActiveProgram"] = self.data["key"]
                return result
        elif self.event == "SELECTED-PROGRAM-REQUEST":
            if self.error_key == "SDK.Error.NoProgramSelected":
                return {"BSH.Common.Root.SelectedProgram": None}
            else:
                result = {item["key"]: item["value"] for item in self.data["options"]}
                result["BSH.Common.Root.SelectedProgram"] = self.data["key"]
                return result
        return {}

    @property
    def error_key(self) -> str | None:
        if self.error:
            return self.error["key"]
        return None

    def __str__(self) -> str:
        data = dict(appliance_id=self.appliance_id, event=self.event, timestamp=self.timestamp)
        if self.data:
            data["data"] = self.data
        if self.error:
            data["error"] = self.error
        return dumps(data) + "\n"

    @property
    def trigger(self) -> Trigger | None:
        if self.event in ("CONNECTED", "PAIRED"):
            # For any new(ly connected) appliance, we perform all requests.
            return Trigger(
                appliance_id=self.appliance_id, status=True, settings=True, selected_program=True, active_program=True
            )
        elif self.event in ("NOTIFY", "EVENT"):
            # For any appliance that is active, request the status once in a while.
            return Trigger(appliance_id=self.appliance_id, status=True, interval=True)
        elif self.event == "STATUS":
            # For an appliance that just switched to running, get the program.
            if self.items.get("BSH.Common.Status.OperationState", "") == "BSH.Common.EnumType.OperationState.Run":
                return Trigger(appliance_id=self.appliance_id, active_program=True, settings=True)
            # For an appliance that was controlled locally, but is no more: get the programs once in a while.
            if self.items.get("BSH.Common.Status.LocalControlActive") is False:
                return Trigger(
                    appliance_id=self.appliance_id, active_program=True, selected_program=True, interval=True
                )
        return None



from time import time
from typing import TYPE_CHECKING

from homeconnect_watcher.event import HomeConnectEvent
from homeconnect_watcher.exceptions import HomeConnectRequestError

if TYPE_CHECKING:
    from .client import HomeConnectClient


class HomeConnectAppliance:
    def __init__(self, client: "HomeConnectClient", appliance_id: str, appliance_type: str):
        self.client = client
        self.appliance_id = appliance_id
        self.appliance_type = appliance_type
        self._available_programs: list[str] | None = None
        self._last_update: float = time()

    def __repr__(self) -> str:
        return f"HomeConnect{self.appliance_type}(ha_id={repr(self.appliance_id)})"

    @property
    def time_since_update(self) -> float:
        return time() - self._last_update

    async def get_available_programs(self) -> list[str] | None:
        if self._available_programs is None:
            response = await self.client._get(f"/{self.appliance_id}/programs/available")
            if "error" in response:
                # Some appliances do not support programs.
                if response["error"]["key"] == "SDK.Error.UnsupportedOperation":
                    self._available_programs = []
                elif response["error"]["key"] in (
                    "SDK.Error.HomeAppliance.Connection.Initialization.Failed",
                    "SDK.Error.WrongOperationState",
                ):
                    return None
                else:
                    self.client.logger.error(response["error"])
                    raise HomeConnectRequestError(response["error"]["key"])
            else:
                self._available_programs = [program["key"] for program in response["data"]["programs"]]
        return self._available_programs if len(self._available_programs) else None

    async def get_status(self) -> HomeConnectEvent:
        self._last_update: float = time()
        response = await self.client._get(f"/{self.appliance_id}/status")
        return HomeConnectEvent.from_request(request="STATUS", appliance_id=self.appliance_id, response=response)

    async def get_settings(self) -> HomeConnectEvent:
        self._last_update: float = time()
        response = await self.client._get(f"/{self.appliance_id}/settings")
        return HomeConnectEvent.from_request(request="SETTINGS", appliance_id=self.appliance_id, response=response)

    async def get_active_program(self) -> HomeConnectEvent:
        self._last_update: float = time()
        response = await self.client._get(f"/{self.appliance_id}/programs/active")
        return HomeConnectEvent.from_request(
            request="ACTIVE-PROGRAM", appliance_id=self.appliance_id, response=response
        )

    async def get_selected_program(self) -> HomeConnectEvent:
        self._last_update: float = time()
        response = await self.client._get(f"/{self.appliance_id}/programs/selected")
        return HomeConnectEvent.from_request(
            request="SELECTED-PROGRAM", appliance_id=self.appliance_id, response=response
        )



class MissingTokenError(Exception):
    pass


class HomeConnectTimeout(Exception):
    pass


class HomeConnectRequestError(Exception):
    def __init__(self, data):
        self.data = data


class HomeConnectParsingError(Exception):
    pass


class HomeConnectConnectionClosed(Exception):
    pass



from asyncio import sleep
from json import load, dump
from logging import getLogger
from os import environ
from pathlib import Path
from re import search
from time import monotonic
from typing import AsyncIterable

from authlib.integrations.httpx_client import AsyncOAuth2Client
from authlib.oauth2.rfc6749.wrappers import OAuth2Token
from requests import Session
from httpx import StreamError, ReadTimeout, RemoteProtocolError


from homeconnect_watcher.client.appliance import HomeConnectAppliance
from homeconnect_watcher.client.trigger import Trigger
from homeconnect_watcher.event import HomeConnectEvent
from homeconnect_watcher.exceptions import HomeConnectRequestError, HomeConnectConnectionClosed, HomeConnectTimeout
from homeconnect_watcher.utils import Metrics, retry, timeout


class HomeConnectClient:
    _authorize_endpoint = "https://api.home-connect.com/security/oauth/authorize"
    _appliances_endpoint = "https://api.home-connect.com/api/homeappliances"
    _token_endpoint = "https://api.home-connect.com/security/oauth/token"

    def __init__(self, token_cache: Path = Path("token"), metrics: Metrics | None = None):
        self.token_cache = (
            Path(environ["HOMECONNECT_PATH"]) / token_cache if "HOMECONNECT_PATH" in environ else token_cache
        )
        self.logger = getLogger(self.__class__.__name__)
        self.client = AsyncOAuth2Client(
            client_id=environ["HOMECONNECT_CLIENT_ID"],
            client_secret=environ["HOMECONNECT_CLIENT_SECRET"],
            redirect_uri=environ["HOMECONNECT_REDIRECT_URI"],
            update_token=self._save_token,
            token_endpoint=self._token_endpoint,
            token=self._load_token(),
        )
        self.client.auth = self.client.token_auth  # Ensure we use token auth for all requests.
        self._appliances: list["HomeConnectAppliance"] | None = None
        self._last_event: int | None = None
        self.metrics = metrics
        if self.metrics:
            self.metrics.set_last_event(lambda: monotonic() - self._last_event if self._last_event else -1)
            self.metrics.set_n_appliances(lambda: len(self._appliances) if self._appliances else 0)

    @property
    async def appliances(self) -> list["HomeConnectAppliance"]:
        if self._appliances is None:
            self._appliances = await self._get_appliances()
        return self._appliances

    async def get_appliance(self, appliance_id: str) -> "HomeConnectAppliance":
        for appliance in await self.appliances:
            if appliance.appliance_id == appliance_id:
                return appliance
        raise KeyError(f"No such appliance: {appliance_id}")

    async def refresh_appliances(self) -> None:
        """Refresh the cached list of appliances. Should only be needed if a new appliance is added."""
        self._appliances = None
        await self.appliances

    async def _get_appliances(self) -> list["HomeConnectAppliance"]:
        data = await self._get("")
        if "error" in data:
            raise HomeConnectRequestError(data["error"])
        return [
            HomeConnectAppliance(
                client=self,
                appliance_id=appliance["haId"],
                appliance_type=appliance["type"],
            )
            for appliance in data["data"]["homeappliances"]
        ]

    @property
    def authorization_url(self) -> str:
        url, state = self.client.create_authorization_url(self._authorize_endpoint)
        return url

    async def authorize(self, url: str) -> None:
        await self.client.fetch_token(self._token_endpoint, authorization_response=url)
        await self._save_token(self.client.token)

    async def events(self, appliance_id: str | None = None) -> AsyncIterable[HomeConnectEvent]:
        """
        Connect to an event stream.

        If appliance_id is provided, will watch only to events of the given appliance.
        Otherwise, will watch to the global event stream.
        """
        url = (
            f"{self._appliances_endpoint}/events"
            if appliance_id is None
            else f"{self._appliances_endpoint}/{appliance_id}/events"
        )
        self.logger.info("Opening event stream.")
        async with self.client.stream("GET", url, timeout=None) as event_stream:
            if event_stream.status_code != 200:
                self.logger.warning(f"Failed to connect to events endpoint. Status code: {event_stream.status_code}")
                raise HomeConnectConnectionClosed()
            try:
                async for entry in timeout(event_stream.aiter_bytes(), duration=120):
                    if entry:
                        yield HomeConnectEvent.from_stream(entry)
                        self._last_event = monotonic()
            except (RemoteProtocolError, StreamError) as e:
                self.logger.warning("Stream error:", exc_info=e)
                raise HomeConnectConnectionClosed()
            except StopAsyncIteration:
                self.logger.info("Reached end of events stream.")
                return

    async def watch(
        self, appliance_id: str | None = None, reconnect_delay: int = 120
    ) -> AsyncIterable[HomeConnectEvent]:
        """
        Watch the status of one or all appliances.

        Listen to the event stream and yield all events.
        """
        for appliance in await self.appliances:
            async for triggered_event in self._handle_trigger(
                Trigger(
                    appliance_id=appliance.appliance_id,
                    status=True,
                    settings=True,
                    active_program=True,
                    selected_program=True,
                )
            ):
                yield triggered_event
                if self.metrics:
                    self.metrics.increment_event_counter(triggered_event)
        while True:
            try:
                async for event in self.events(appliance_id=appliance_id):
                    yield event
                    if self.metrics:
                        self.metrics.increment_event_counter(event)
                    async for triggered_event in self._handle_trigger(event.trigger):
                        yield triggered_event
                        if self.metrics:
                            self.metrics.increment_event_counter(triggered_event)
            except HomeConnectConnectionClosed:
                self.logger.warning(f"Connection closed. Reconnecting in {reconnect_delay} seconds.")
                if self.metrics:
                    self.metrics.increment_disconnects(reason="closed")
                await sleep(reconnect_delay)
                continue
            except HomeConnectTimeout:
                self.logger.warning(f"Connection timed out. Reconnecting in {reconnect_delay} seconds.")
                if self.metrics:
                    self.metrics.increment_disconnects(reason="timeout")
                await sleep(reconnect_delay)
                continue

    def _load_token(self) -> OAuth2Token | None:
        """Load the OAuth token from file."""
        if not self.token_cache.is_file():
            self.logger.warning("Token cache not found.")
            return
        with self.token_cache.open() as token_file:
            self.logger.info("Loading token from disk.")
            return OAuth2Token(load(token_file))

    async def _save_token(self, token: OAuth2Token, refresh_token=None) -> None:  # noqa
        """Save the OAuth token to the token cache."""
        if self.metrics:
            self.metrics.increment_token_refresh()
        with self.token_cache.open(mode="w") as token_file:
            self.logger.info("Saving token to disk.")
            dump(token, token_file)

    @retry(n_tries=3, exceptions=(ReadTimeout,))
    async def _get(self, path: str) -> dict[str, ...]:
        await sleep(delay=1.5)  # Rate limit
        resp = await self.client.get(f"{self._appliances_endpoint}{path}")
        data = resp.json()
        if len(data.keys()) > 1:
            raise KeyError(f"Unexpected keys: {data.keys()}")
        return data

    async def _handle_trigger(self, trigger: Trigger | None) -> AsyncIterable[HomeConnectEvent]:
        if trigger is None:
            return
        appliance = await self.get_appliance(trigger.appliance_id)
        if trigger.interval and appliance.time_since_update < 300:
            return  # If the trigger is an interval trigger, only do requests if last requests were 5 minutes ago.
        if trigger.status:
            yield await appliance.get_status()
        if trigger.settings:
            yield await appliance.get_settings()
        if await appliance.get_available_programs():  # Only if the appliance supports programs.
            if trigger.active_program:
                yield await appliance.get_active_program()
            if trigger.selected_program:
                yield await appliance.get_selected_program()


class HomeConnectSimulationClient(HomeConnectClient):
    _authorize_endpoint = "https://simulator.home-connect.com/security/oauth/authorize"
    _appliances_endpoint = "https://simulator.home-connect.com/api/homeappliances"
    _token_endpoint = "https://simulator.home-connect.com/security/oauth/token"

    def __init__(self, token_cache: Path = Path("simulation_token"), metrics: Metrics | None = None):
        super().__init__(token_cache=token_cache, metrics=metrics)

    async def authenticate(self, username: str, password: str):
        """Automatic simulation login. Any username/password goes, except 'wrongPassword'."""
        session = Session()
        response = session.post(
            self.authorization_url.replace("authorize", "login"),
            data=dict(email=username, password=password),
        )
        grant_url = search('(/security/oauth/grant.*)" +method="post"', str(response.content))
        if grant_url is None:
            raise ValueError(response.content)
        response = session.post(
            "https://simulator.home-connect.com" + grant_url.group(1),
            data={
                "submit": "approve",
                "user": "email",
                "client_id": environ["HOMECONNECT_CLIENT_ID"],
                "scope": "CleaningRobot CleaningRobot-Control CleaningRobot-Monitor CleaningRobot-Settings CoffeeMaker "
                "CoffeeMaker-Control CoffeeMaker-Monitor CoffeeMaker-Settings Control CookProcessor "
                "CookProcessor-Control CookProcessor-Monitor CookProcessor-Settings Dishwasher "
                "Dishwasher-Control Dishwasher-Monitor Dishwasher-Settings Dryer Dryer-Control Dryer-Monitor "
                "Dryer-Settings Freezer Freezer-Control Freezer-Monitor Freezer-Settings "
                "FridgeFreezer-Control FridgeFreezer-Monitor FridgeFreezer-Settings Hob Hob-Control "
                "Hob-Monitor Hob-Settings Hood Hood-Control Hood-Monitor Hood-Settings IdentifyAppliance "
                "Monitor Oven Oven-Control Oven-Monitor Oven-Settings Refrigerator Refrigerator-Control "
                "Refrigerator-Monitor Refrigerator-Settings Settings Washer Washer-Control Washer-Monitor "
                "Washer-Settings WasherDryer WasherDryer-Control WasherDryer-Monitor WasherDryer-Settings "
                "WineCooler WineCooler-Control WineCooler-Monitor WineCooler-Settings",
                "redirect_uri": environ["HOMECONNECT_REDIRECT_URI"],
            },
            allow_redirects=False,
        )
        response = session.get(response.headers["Location"], allow_redirects=False)
        await self.authorize(url=response.headers["Location"])



from abc import ABCMeta, abstractmethod
from logging import getLogger


from homeconnect_watcher.event import HomeConnectEvent


class BaseExporter(metaclass=ABCMeta):
    def __init__(self):
        self.logger = getLogger(self.__class__.__name__)

    def __enter__(self) -> "BaseExporter":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        return

    @abstractmethod
    def export(self, event: HomeConnectEvent) -> None:
        pass



from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional, TextIO

from homeconnect_watcher.event import HomeConnectEvent

from .base import BaseExporter


class FileExporter(BaseExporter):
    def __init__(self, path: Path, flush_interval: timedelta = timedelta(minutes=30)):
        super().__init__()
        self.path = path
        self.flush_interval = flush_interval
        self._fp: Optional[TextIO] = None
        self._last_flush: datetime = datetime.now()

    def __enter__(self) -> "FileExporter":
        self._fp = self._open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._fp.close()
        self._fp = None
        return

    def export(self, event: HomeConnectEvent) -> None:
        self._fp.write(str(event))
        self._flush()

    def _flush(self) -> None:
        now = datetime.now()
        if now.date() != self._last_flush.date():
            self._fp.close()
            self._fp = self._open()
        elif now - self._last_flush > self.flush_interval:
            self.logger.info(f"Flushing output file.")
            self._last_flush = datetime.now()
            self._fp.flush()

    def _open(self) -> TextIO:
        self._last_flush = datetime.now()
        path = self.path / f"hcw_{self._last_flush.date().strftime('%Y-%m-%d')}.jsonl"
        self.logger.info(f"Opening output file {str(path)}.")
        return path.open("a")



from asyncio import run as async_run
from typing import Optional

from fastapi import FastAPI
from fastapi.responses import Response
from typer import Argument, Typer
from uvicorn import run

from homeconnect_watcher.api import loop
from homeconnect_watcher.client.client import HomeConnectSimulationClient, HomeConnectClient
from homeconnect_watcher.utils import LogLevel, Metrics, initialize_logging

app = Typer()


@app.command()
def authorize(
    log_level: LogLevel = Argument(LogLevel.INFO, envvar="HCW_LOGLEVEL"),
):
    initialize_logging(level=log_level)
    client = HomeConnectClient()

    api = FastAPI()

    @api.get("/code")
    async def code(code: str):
        url = f"{client.client.redirect_uri}/?code={code}&grant_type=authorization_code"
        await client.authorize(url)
        print("Authorized.")
        return Response("OK", 200)

    print("Visit the following URL:")
    print(client.authorization_url)
    run(api, host="0.0.0.0", port=8000)


@app.command()
def watch(
    simulation: bool = False,
    flush_interval: int = Argument(300, envvar="HCW_FLUSH_INTERVAL"),
    log_level: LogLevel = Argument(LogLevel.INFO, envvar="HCW_LOGLEVEL"),
    metrics_port: Optional[int] = Argument(None, envvar="HCW_METRICS_PORT"),
):
    initialize_logging(level=log_level)
    metrics = Metrics(port=metrics_port) if metrics_port else None
    client = (HomeConnectSimulationClient if simulation else HomeConnectClient)(metrics=metrics)
    async_run(loop(client, flush_interval=flush_interval))



from asyncio import create_task, Task
from datetime import timedelta
from logging import getLogger
from os import environ
from pathlib import Path
from time import time
from typing import Optional, Type

from dotenv import load_dotenv
from fastapi import FastAPI
from fastapi.responses import Response, RedirectResponse

from homeconnect_watcher.client.client import HomeConnectClient, HomeConnectSimulationClient
from homeconnect_watcher.exceptions import HomeConnectRequestError
from homeconnect_watcher.exporter.file import FileExporter

load_dotenv()


def define_endpoints(simulation: bool):
    logger = getLogger(__name__)

    api = FastAPI()
    client: HomeConnectClient | None = None
    task: Task | None = None

    @api.on_event("startup")
    async def startup():
        nonlocal client, task
        if task is not None:
            task.cancel()
        if client is not None:
            await client.client.aclose()
        client = (HomeConnectSimulationClient if simulation else HomeConnectClient)()
        if simulation:
            await client.authenticate("user", "password")
        task = create_task(loop(client))

    @api.on_event("shutdown")
    async def shutdown():
        await client.client.aclose()

    @api.get("/authorize")
    async def authorize():
        return RedirectResponse(client.authorization_url)

    @api.get("/code")
    async def code(code: str):
        url = f"{client.client.redirect_uri}/?code={code}&grant_type=authorization_code"
        await client.authorize(url)
        await startup()
        return Response("OK", 200)

    @api.get("/health")
    async def health():
        if task is None:
            logger.warning("No task running.")
            return Response("No task running.", status_code=404)
        elif task.done():
            exc = task.exception()
            if "invalid_token" in str(exc):
                logger.warning("Unauthorized. Please authorize.")
                return Response("Unauthorized. Please authorize.")
            logger.warning("Task finished.", exc_info=exc)
            return Response(f"Task finished: {exc}", status_code=500)
        else:
            # TODO: last events
            age = 0  # int(time() - client.last_event)
            if age > 240:
                logger.warning(f"No recent events. Last event {age} seconds old.")
                return Response(f"No recent events. Last event {age} seconds old.", status_code=500)
            else:
                return Response(f"OK. Last event {age} seconds old.", status_code=200)

    return api


async def loop(client: HomeConnectClient, flush_interval: int):
    if isinstance(client, HomeConnectSimulationClient):
        await client.authenticate("username", "password")
    path = Path(environ["HOMECONNECT_PATH"]) if "HOMECONNECT_PATH" in environ else Path()
    exporter = FileExporter(path, flush_interval=timedelta(seconds=flush_interval))
    with exporter:
        async for event in client.watch():
            exporter.export(event)
